@@ -54,6 +54,11 @@ static void update_focus_mode      (MetaScreen *screen);
 static void prefs_changed_callback (MetaPreference pref,
                                     gpointer       data);
 
+#ifdef HAVE_STARTUP_NOTIFICATION
+static void meta_screen_sn_event   (SnMonitorEvent *event,
+                                    void           *user_data);
+#endif
+
 static int
 set_wm_check_hint (MetaScreen *screen)
 {
@@ -404,11 +409,11 @@ meta_screen_new (MetaDisplay *display,
         }
 
       /* We want to find out when the current selection owner dies */
-      meta_error_trap_push (display);
+      meta_error_trap_push_with_return (display);
       attrs.event_mask = StructureNotifyMask;
       XChangeWindowAttributes (xdisplay,
                                current_wm_sn_owner, CWEventMask, &attrs);
-      if (meta_error_trap_pop (display) != Success)
+      if (meta_error_trap_pop_with_return (display, FALSE) != Success)
         current_wm_sn_owner = None; /* don't wait for it to die later on */
     }
 
@@ -477,7 +482,7 @@ meta_screen_new (MetaDisplay *display,
     }
   
   /* select our root window events */
-  meta_error_trap_push (display);
+  meta_error_trap_push_with_return (display);
 
   /* We need to or with the existing event mask since
    * gtk+ may be interested in other events.
@@ -492,7 +497,7 @@ meta_screen_new (MetaDisplay *display,
                 KeyPressMask | KeyReleaseMask |
                 FocusChangeMask | StructureNotifyMask |
 		attr.your_event_mask);
-  if (meta_error_trap_pop (display) != Success)
+  if (meta_error_trap_pop_with_return (display, FALSE) != Success)
     {
       meta_warning (_("Screen %d on display \"%s\" already has a window manager\n"),
                     number, display->name);
@@ -523,15 +528,18 @@ meta_screen_new (MetaDisplay *display,
   screen->work_area_idle = 0;
 
   screen->active_workspace = NULL;
+  screen->workspaces = NULL;
   screen->rows_of_workspaces = 1;
   screen->columns_of_workspaces = -1;
   screen->vertical_workspaces = FALSE;
   screen->starting_corner = META_SCREEN_TOPLEFT;
 
+  screen->showing_desktop = FALSE;
+  
   screen->xinerama_infos = NULL;
   screen->n_xinerama_infos = 0;
   screen->last_xinerama_index = 0;
-
+  
   reload_xinerama_infos (screen);
   
   meta_screen_set_cursor (screen, META_CURSOR_DEFAULT);
@@ -577,6 +585,17 @@ meta_screen_new (MetaDisplay *display,
   screen->stack = meta_stack_new (screen);
 
   meta_prefs_add_listener (prefs_changed_callback, screen);
+
+#ifdef HAVE_STARTUP_NOTIFICATION
+  screen->sn_context =
+    sn_monitor_context_new (screen->display->sn_display,
+                            screen->number,
+                            meta_screen_sn_event,
+                            screen,
+                            NULL);
+  screen->startup_sequences = NULL;
+  screen->startup_sequence_timeout = 0;
+#endif
   
   meta_verbose ("Added screen %d ('%s') root 0x%lx\n",
                 screen->number, screen->screen_name, screen->xroot);  
@@ -601,13 +620,31 @@ meta_screen_free (MetaScreen *screen)
   
   meta_screen_ungrab_keys (screen);
 
+#ifdef HAVE_STARTUP_NOTIFICATION
+  g_slist_foreach (screen->startup_sequences,
+                   (GFunc) sn_startup_sequence_unref, NULL);
+  g_slist_free (screen->startup_sequences);
+  screen->startup_sequences = NULL;
+
+  if (screen->startup_sequence_timeout != 0)
+    {
+      g_source_remove (screen->startup_sequence_timeout);
+      screen->startup_sequence_timeout = 0;
+    }
+  if (screen->sn_context)
+    {
+      sn_monitor_context_unref (screen->sn_context);
+      screen->sn_context = NULL;
+    }
+#endif
+  
   meta_ui_free (screen->ui);
 
   meta_stack_free (screen->stack);
 
-  meta_error_trap_push (screen->display);
+  meta_error_trap_push_with_return (screen->display);
   XSelectInput (screen->display->xdisplay, screen->xroot, 0);
-  if (meta_error_trap_pop (screen->display) != Success)
+  if (meta_error_trap_pop_with_return (screen->display, FALSE) != Success)
     meta_warning (_("Could not release screen %d on display \"%s\"\n"),
                   screen->number, screen->display->name);
 
@@ -635,13 +672,13 @@ meta_screen_manage_all_windows (MetaScreen *screen)
   /* Must grab server to avoid obvious race condition */
   meta_display_grab (screen->display);
 
-  meta_error_trap_push (screen->display);
+  meta_error_trap_push_with_return (screen->display);
   
   XQueryTree (screen->display->xdisplay,
               screen->xroot,
               &ignored1, &ignored2, &children, &n_children);
 
-  if (meta_error_trap_pop (screen->display))
+  if (meta_error_trap_pop_with_return (screen->display, TRUE) != Success)
     {
       meta_display_ungrab (screen->display);
       return;
@@ -810,33 +847,45 @@ meta_screen_queue_window_resizes (MetaScreen *screen)
 
 int
 meta_screen_get_n_workspaces (MetaScreen *screen)
+{
+  return g_list_length (screen->workspaces);
+}
+
+MetaWorkspace*
+meta_screen_get_workspace_by_index (MetaScreen  *screen,
+                                    int          idx)
 {
   GList *tmp;
   int i;
 
+  /* should be robust, idx is maybe from an app */
+  if (idx < 0)
+    return NULL;
+  
   i = 0;
-  tmp = screen->display->workspaces;
+  tmp = screen->workspaces;
   while (tmp != NULL)
     {
       MetaWorkspace *w = tmp->data;
 
-      if (w->screen == screen)
-        ++i;
-      
+      if (i == idx)
+        return w;
+
+      ++i;
       tmp = tmp->next;
     }
 
-  return i;
+  return NULL;
 }
 
-static int
+static void
 set_number_of_spaces_hint (MetaScreen *screen,
 			   int         n_spaces)
 {
   unsigned long data[1];
 
   if (screen->closing > 0)
-    return 0;
+    return;
 
   data[0] = n_spaces;
 
@@ -847,7 +896,7 @@ set_number_of_spaces_hint (MetaScreen *screen,
                    screen->display->atom_net_number_of_desktops,
                    XA_CARDINAL,
                    32, PropModeReplace, (guchar*) data, 1);
-  return meta_error_trap_pop (screen->display);
+  meta_error_trap_pop (screen->display, FALSE);
 }
 
 static void
@@ -867,21 +916,17 @@ update_num_workspaces (MetaScreen *screen)
   last_remaining = NULL;
   extras = NULL;
   i = 0;
-  tmp = screen->display->workspaces;
+  tmp = screen->workspaces;
   while (tmp != NULL)
     {
       MetaWorkspace *w = tmp->data;
 
-      if (w->screen == screen)
-        {          
-          ++i;
-
-          if (i > new_num)
-            extras = g_list_prepend (extras, w);
-          else
-            last_remaining = w;
-        }
-      
+      if (i > new_num)
+        extras = g_list_prepend (extras, w);
+      else
+        last_remaining = w;
+          
+      ++i;
       tmp = tmp->next;
     }
 
@@ -994,7 +1039,10 @@ meta_screen_ensure_tab_popup (MetaScreen *screen,
       entries[i].key = (MetaTabEntryKey) window->xwindow;
       entries[i].title = window->title;
       entries[i].icon = window->icon;
-      meta_window_get_outer_rect (window, &r);
+
+      if (!window->minimized || !meta_window_get_icon_geometry (window, &r))
+        meta_window_get_outer_rect (window, &r);
+      
       entries[i].x = r.x;
       entries[i].y = r.y;
       entries[i].width = r.width;
@@ -1078,7 +1126,7 @@ meta_screen_ensure_workspace_popup (MetaScreen *screen)
               if (k >= len)
                 break;
 
-              workspace = meta_display_get_workspace_by_index (screen->display, k);
+              workspace = meta_screen_get_workspace_by_index (screen, k);
               g_assert (workspace);
 
               entries[iter].key = (MetaTabEntryKey) workspace;
@@ -1096,7 +1144,7 @@ meta_screen_ensure_workspace_popup (MetaScreen *screen)
         {
           MetaWorkspace *workspace;
 
-          workspace = meta_display_get_workspace_by_index (screen->display, i);
+          workspace = meta_screen_get_workspace_by_index (screen, i);
 
           g_assert (workspace);
 
@@ -1362,17 +1410,14 @@ meta_screen_update_workspace_names (MetaScreen *screen)
     }
 
   i = 0;
-  tmp = screen->display->workspaces;
+  tmp = screen->workspaces;
   while (tmp != NULL && i < n_names)
     {
       MetaWorkspace *w = tmp->data;
 
-      if (w->screen == screen)
-        {
-          meta_workspace_set_name (w, names[i]);
+      meta_workspace_set_name (w, names[i]);
           
-          ++i;
-        }
+      ++i;
       
       tmp = tmp->next;
     }
@@ -1403,7 +1448,7 @@ meta_create_offscreen_window (Display *xdisplay,
                         &attrs);
 }
 
-static int
+static void
 set_work_area_hint (MetaScreen *screen)
 {
   int num_workspaces;
@@ -1413,7 +1458,7 @@ set_work_area_hint (MetaScreen *screen)
   
   num_workspaces = meta_screen_get_n_workspaces (screen);
   data = g_new (unsigned long, num_workspaces * 4);
-  tmp_list = screen->display->workspaces;
+  tmp_list = screen->workspaces;
   tmp = data;
   
   while (tmp_list != NULL)
@@ -1440,7 +1485,7 @@ set_work_area_hint (MetaScreen *screen)
 		   XA_CARDINAL, 32, PropModeReplace,
 		   (guchar*) data, num_workspaces*4);
   g_free (data);
-  return meta_error_trap_pop (screen->display);
+  meta_error_trap_pop (screen->display, FALSE);
 }
 
 static gboolean
@@ -1543,3 +1588,252 @@ meta_screen_resize (MetaScreen *screen,
   /* Queue a resize on all the windows */
   meta_screen_foreach_window (screen, meta_screen_resize_func, 0);
 }
+
+static void
+update_showing_desktop_hint (MetaScreen *screen)
+{
+  unsigned long data[1];
+
+  data[0] = screen->showing_desktop ? 1 : 0;  
+      
+  meta_error_trap_push (screen->display);
+  XChangeProperty (screen->display->xdisplay, screen->xroot,
+                   screen->display->atom_net_showing_desktop,
+                   XA_CARDINAL,
+                   32, PropModeReplace, (guchar*) data, 1);
+  meta_error_trap_pop (screen->display, FALSE);
+}
+
+static void
+queue_windows_showing (MetaScreen *screen)
+{
+  GSList *windows;
+  GSList *tmp;
+
+  windows = meta_display_list_windows (screen->display);
+
+  tmp = windows;
+  while (tmp != NULL)
+    {
+      MetaWindow *w = tmp->data;
+
+      if (w->screen == screen)
+        meta_window_queue_calc_showing (w);
+      
+      tmp = tmp->next;
+    }
+
+  g_slist_free (windows);
+}
+
+void
+meta_screen_show_desktop (MetaScreen *screen)
+{
+  if (screen->showing_desktop)
+    return;
+
+  screen->showing_desktop = TRUE;
+
+  queue_windows_showing (screen);
+
+  update_showing_desktop_hint (screen);
+}
+
+void
+meta_screen_unshow_desktop (MetaScreen *screen)
+{
+  if (!screen->showing_desktop)
+    return;
+
+  screen->showing_desktop = FALSE;
+  
+  queue_windows_showing (screen);
+
+  update_showing_desktop_hint (screen);
+
+  meta_screen_focus_top_window (screen, NULL);
+}
+
+#ifdef HAVE_STARTUP_NOTIFICATION
+static gboolean startup_sequence_timeout (void *data);
+
+static void
+update_startup_feedback (MetaScreen *screen)
+{
+  if (screen->startup_sequences != NULL)
+    {
+      meta_topic (META_DEBUG_STARTUP,
+                  "Setting busy cursor\n");
+      meta_screen_set_cursor (screen, META_CURSOR_BUSY);
+    }
+  else
+    {
+      meta_topic (META_DEBUG_STARTUP,
+                  "Setting default cursor\n");
+      meta_screen_set_cursor (screen, META_CURSOR_DEFAULT);
+    }
+}
+
+static void
+add_sequence (MetaScreen        *screen,
+              SnStartupSequence *sequence)
+{
+  meta_topic (META_DEBUG_STARTUP,
+              "Adding sequence %s\n",
+              sn_startup_sequence_get_id (sequence));
+  sn_startup_sequence_ref (sequence);
+  screen->startup_sequences = g_slist_prepend (screen->startup_sequences,
+                                               sequence);
+
+  /* our timeout just polls every second, instead of bothering
+   * to compute exactly when we may next time out
+   */
+  if (screen->startup_sequence_timeout == 0)
+    screen->startup_sequence_timeout = g_timeout_add (1000,
+                                                      startup_sequence_timeout,
+                                                      screen);
+
+  update_startup_feedback (screen);
+}
+
+static void
+remove_sequence (MetaScreen        *screen,
+                 SnStartupSequence *sequence)
+{
+  meta_topic (META_DEBUG_STARTUP,
+              "Removing sequence %s\n",
+              sn_startup_sequence_get_id (sequence));
+  
+  screen->startup_sequences = g_slist_remove (screen->startup_sequences,
+                                              sequence);
+  sn_startup_sequence_unref (sequence);
+  
+  if (screen->startup_sequences == NULL &&
+      screen->startup_sequence_timeout != 0)
+    {
+      g_source_remove (screen->startup_sequence_timeout);
+      screen->startup_sequence_timeout = 0;
+    }
+
+  update_startup_feedback (screen);
+}
+
+typedef struct
+{
+  GSList *list;
+  GTimeVal now;
+} CollectTimedOutData;
+
+#define STARTUP_TIMEOUT 5000
+
+static void
+collect_timed_out_foreach (void *element,
+                           void *data)
+{
+  CollectTimedOutData *ctod = data;
+  SnStartupSequence *sequence = element;
+  long tv_sec, tv_usec;
+  double elapsed;
+  
+  sn_startup_sequence_get_last_active_time (sequence, &tv_sec, &tv_usec);
+
+  elapsed =
+    ((((double)ctod->now.tv_sec - tv_sec) * G_USEC_PER_SEC +
+      (ctod->now.tv_usec - tv_usec))) / 1000.0;
+
+  meta_topic (META_DEBUG_STARTUP,
+              "Sequence used %g seconds vs. %g max: %s\n",
+              elapsed, (double) STARTUP_TIMEOUT,
+              sn_startup_sequence_get_id (sequence));
+  
+  if (elapsed > STARTUP_TIMEOUT)
+    ctod->list = g_slist_prepend (ctod->list, sequence);
+}
+
+static gboolean
+startup_sequence_timeout (void *data)
+{
+  MetaScreen *screen = data;
+  CollectTimedOutData ctod;
+  GSList *tmp;
+  
+  ctod.list = NULL;
+  g_get_current_time (&ctod.now);
+  g_slist_foreach (screen->startup_sequences,
+                   collect_timed_out_foreach,
+                   &ctod);
+
+  tmp = ctod.list;
+  while (tmp != NULL)
+    {
+      SnStartupSequence *sequence = tmp->data;
+
+      meta_topic (META_DEBUG_STARTUP,
+                  "Timed out sequence %s\n",
+                  sn_startup_sequence_get_id (sequence));
+      
+      remove_sequence (screen, sequence);
+      
+      tmp = tmp->next;
+    }
+
+  g_slist_free (ctod.list);
+  
+  if (screen->startup_sequences != NULL)
+    {
+      return TRUE;
+    }
+  else
+    {
+      /* remove */
+      screen->startup_sequence_timeout = 0;
+      return FALSE;
+    }
+}
+
+static void
+meta_screen_sn_event (SnMonitorEvent *event,
+                      void           *user_data)
+{
+  MetaScreen *screen;
+  SnStartupSequence *sequence;
+  
+  screen = user_data;
+
+  sequence = sn_monitor_event_get_startup_sequence (event);
+  
+  switch (sn_monitor_event_get_type (event))
+    {
+    case SN_MONITOR_EVENT_INITIATED:
+      {
+        meta_topic (META_DEBUG_STARTUP,
+                    "Received startup initiated for %s\n",
+                    sn_startup_sequence_get_id (sequence));
+        add_sequence (screen, sequence);
+      }
+      break;
+
+    case SN_MONITOR_EVENT_COMPLETED:
+      {
+        meta_topic (META_DEBUG_STARTUP,
+                    "Received startup completed for %s\n",
+                    sn_startup_sequence_get_id (sequence));
+        remove_sequence (screen,
+                         sn_monitor_event_get_startup_sequence (event));
+      }
+      break;
+
+    case SN_MONITOR_EVENT_CHANGED:
+      meta_topic (META_DEBUG_STARTUP,
+                  "Received startup changed for %s\n",
+                  sn_startup_sequence_get_id (sequence));
+      break;
+
+    case SN_MONITOR_EVENT_CANCELED:
+      meta_topic (META_DEBUG_STARTUP,
+                  "Received startup canceled for %s\n",
+                  sn_startup_sequence_get_id (sequence));
+      break;
+    }
+}
+#endif
